package ru.dbotthepony.mc.otm.android;

import net.minecraft.ChatFormatting;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.Items;
import net.minecraftforge.registries.RegistryManager;
import ru.dbotthepony.mc.otm.registry.Registry;
import ru.dbotthepony.mc.otm.client.render.SkinElement;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;

public class AndroidResearchBuilder {
	record DeferredItemStack(ResourceLocation id, int amount) {
	}

	record DeferredFeature(ResourceLocation id, int level, Consumer<AndroidFeature> callback) {
		DeferredFeature(ResourceLocation id, int level, Consumer<AndroidFeature> callback) {
			this.id = id;
			this.level = level;
			this.callback = callback;
		}

		DeferredFeature(ResourceLocation id, int level) {
			this(id, level, (f) -> {});
		}

		DeferredFeature(ResourceLocation id, Consumer<AndroidFeature> callback) {
			this(id, 0, (f) -> {});
		}

		DeferredFeature(ResourceLocation id) {
			this(id, 0, (f) -> {});
		}
	}

	public int experience = 0;
	public Component text;
	public boolean has_description = false;
	public Component custom_description;
	public Component custom_name;
	public final ArrayList<DeferredItemStack> items = new ArrayList<>();
	public final ArrayList<ResourceLocation> prerequisites = new ArrayList<>();
	public final ArrayList<DeferredFeature> feature_results = new ArrayList<>();
	public SkinElement icon;

	public AndroidResearchBuilder addPrerequisite(ResourceLocation location) {
		prerequisites.add(location);
		return this;
	}

	public AndroidResearchBuilder addPrerequisite(AndroidResearchType<?> location) {
		prerequisites.add(Objects.requireNonNull(location.getRegistryName()));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(ResourceLocation location) {
		feature_results.add(new DeferredFeature(location));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(AndroidFeatureType<?> location) {
		feature_results.add(new DeferredFeature(Objects.requireNonNull(location.getRegistryName())));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(ResourceLocation location, Consumer<AndroidFeature> callback) {
		feature_results.add(new DeferredFeature(location, callback));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(AndroidFeatureType<?> location, Consumer<AndroidFeature> callback) {
		feature_results.add(new DeferredFeature(Objects.requireNonNull(location.getRegistryName()), callback));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(ResourceLocation location, int level) {
		feature_results.add(new DeferredFeature(location, level));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(AndroidFeatureType<?> location, int level) {
		feature_results.add(new DeferredFeature(Objects.requireNonNull(location.getRegistryName()), level));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(ResourceLocation location, int level, Consumer<AndroidFeature> callback) {
		feature_results.add(new DeferredFeature(location, level, callback));
		return this;
	}

	public AndroidResearchBuilder addFeatureResult(AndroidFeatureType<?> location, int level, Consumer<AndroidFeature> callback) {
		feature_results.add(new DeferredFeature(Objects.requireNonNull(location.getRegistryName()), level, callback));
		return this;
	}

	public AndroidResearchBuilder addItem(ResourceLocation location) {
		items.add(new DeferredItemStack(location, 1));
		return this;
	}

	public AndroidResearchBuilder addItem(ResourceLocation location, int amount) {
		items.add(new DeferredItemStack(location, amount));
		return this;
	}

	public AndroidResearchBuilder setExperienceCost(int experience) {
		this.experience = experience;
		return this;
	}

	public AndroidResearchBuilder withDescription() {
		this.has_description = true;
		return this;
	}

	public AndroidResearchBuilder withDescription(Component text) {
		this.has_description = true;
		this.custom_description = text;
		return this;
	}

	public AndroidResearchBuilder withName(Component text) {
		this.custom_name = text;
		return this;
	}

	public AndroidResearchBuilder withIcon(SkinElement icon) {
		this.icon = icon;
		return this;
	}

	public AndroidResearchBuilder withIconText(Component text) {
		this.text = text;
		return this;
	}

	private final ArrayList<ItemStack> resolved_stacks = new ArrayList<>();
	private List<AndroidResearchType<?>> resolved_preq;

	record ResolvedFeature(AndroidFeatureType<?> type, int level, Consumer<AndroidFeature> callback) {
	}

	private final ArrayList<ResolvedFeature> resolved_features = new ArrayList<>();

	private boolean resolved = false;

	private static int getXpNeededForLevel(int level) {
		int accumulated = 0;

		for (int thisLevel = 0; thisLevel < level; thisLevel++) {
			if (thisLevel >= 30) {
				accumulated += 112 + (thisLevel - 30) * 9;
			} else {
				accumulated += thisLevel >= 15 ? 37 + (thisLevel - 15) * 5 : 7 + thisLevel * 2;
			}
		}

		return accumulated;
	}

	private void resolve() {
		if (!resolved) {
			resolved = true;

			for (var item : items) {
				var get_item = RegistryManager.ACTIVE.getRegistry(Item.class).getValue(item.id);

				if (get_item != null && get_item != Items.AIR) {
					resolved_stacks.add(new ItemStack(get_item, item.amount));
				}
			}

			var list = new ArrayList<AndroidResearchType<?>>();

			for (var entry : prerequisites) {
				var get = Registry.ANDROID_RESEARCH.getValue(entry);

				if (get != null) {
					list.add(get);
				} else {
					throw new IllegalArgumentException("Can not find android research " + entry);
				}
			}

			resolved_preq = List.copyOf(list);

			for (var entry : feature_results) {
				var get = Registry.ANDROID_FEATURES.getValue(entry.id);

				if (get != null) {
					resolved_features.add(new ResolvedFeature(get, entry.level, entry.callback));
				} else {
					throw new IllegalArgumentException("Can not find android feature " + entry);
				}
			}
		}
	}

	public AndroidResearchType<AndroidResearch> build() {
		return new AndroidResearchType<>((type, capability) -> {
			resolve();

			return new AndroidResearch(type, capability) {
				@Nullable
				@Override
				public SkinElement getIcon() {
					return icon;
				}

				@Nullable
				@Override
				public Component getIconText() {
					return text;
				}

				@Override
				public List<Component> getTooltip() {
					var list = new ArrayList<>(super.getTooltip());

					if (custom_description != null) {
						list.add(custom_description);
					} else if (has_description) {
						list.add(new TranslatableComponent("android_research." + type.getRegistryName().getNamespace() + "." + type.getRegistryName().getPath() + ".description"));
					}

					if (!researched) {
						if (experience > 0) {
							if (capability.ply.experienceLevel >= experience) {
								list.add(new TranslatableComponent("otm.android_station.research.xp_cost", experience).withStyle(ChatFormatting.DARK_GREEN));
							} else {
								list.add(new TranslatableComponent("otm.android_station.research.xp_cost", experience).withStyle(ChatFormatting.DARK_RED));
							}
						}

						for (var stack : resolved_stacks) {
							boolean hit = false;

							if (stack.getTag() == null) {
								for (var inv_stack : capability.ply.getInventory().items) {
									if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount()) {
										hit = true;
										break;
									}
								}
							} else {
								for (var inv_stack : capability.ply.getInventory().items) {
									if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount() && inv_stack.getTag() != null && inv_stack.getTag().equals(stack.getTag())) {
										hit = true;
										break;
									}
								}
							}

							if (hit) {
								list.add(new TranslatableComponent("otm.android_station.research.item", stack.getDisplayName(), stack.getCount()).withStyle(ChatFormatting.DARK_GREEN));
							} else {
								list.add(new TranslatableComponent("otm.android_station.research.item", stack.getDisplayName(), stack.getCount()).withStyle(ChatFormatting.DARK_RED));
							}
						}

						for (var research : getPrerequisites()) {
							list.add(new TranslatableComponent("android_research.status.requires", research.getDisplayName()).withStyle(capability.getResearch(research).isResearched() ? ChatFormatting.DARK_GREEN : ChatFormatting.DARK_RED));
						}
					}

					for (var research : getBlocking()) {
						list.add(new TranslatableComponent("android_research.status.blocks", research.getDisplayName()).withStyle(ChatFormatting.DARK_RED));
					}

					for (var research : getBlockedBy()) {
						list.add(new TranslatableComponent("android_research.status.blocked_by", research.getDisplayName()).withStyle(ChatFormatting.DARK_RED));
					}

					return list;
				}

				@Override
				public void onResearched() {
					for (var feature : resolved_features) {
						var get = capability.getFeature(feature.type);

						if (get == null) {
							get = capability.addFeature(feature.type);
							get.setLevel(feature.level);
							feature.callback.accept(get);
						} else {
							if (get.getLevel() < feature.level)
								get.setLevel(feature.level);

							feature.callback.accept(get);
						}
					}
				}

				@Override
				public void onUnresearched() {

				}

				@Override
				public boolean canAfford() {
					if (capability.ply.experienceLevel < experience)
						return false;

					for (var stack : resolved_stacks) {
						boolean hit = false;

						if (stack.getTag() == null) {
							for (var inv_stack : capability.ply.getInventory().items) {
								if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount()) {
									hit = true;
									break;
								}
							}
						} else {
							for (var inv_stack : capability.ply.getInventory().items) {
								if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount() && inv_stack.getTag() != null && inv_stack.getTag().equals(stack.getTag())) {
									hit = true;
									break;
								}
							}
						}

						if (!hit)
							return false;
					}

					return true;
				}

				@Override
				public void refund() {
					capability.ply.giveExperiencePoints(getXpNeededForLevel(experience));

					for (var stack : resolved_stacks) {
						capability.ply.spawnAtLocation(stack);
					}
				}

				@Override
				public void consumeCost() {
					if (experience > 0)
						capability.ply.giveExperienceLevels(-experience);

					for (var stack : resolved_stacks) {
						if (stack.getTag() == null) {
							for (var inv_stack : capability.ply.getInventory().items) {
								if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount()) {
									inv_stack.shrink(stack.getCount());
									break;
								}
							}
						} else {
							for (var inv_stack : capability.ply.getInventory().items) {
								if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount() && inv_stack.getTag() != null && inv_stack.getTag().equals(stack.getTag())) {
									inv_stack.shrink(stack.getCount());
									break;
								}
							}
						}
					}

					capability.ply.getInventory().setChanged();
				}
			};
		}) {
			@Override
			public List<AndroidResearchType<?>> getPrerequisites() {
				resolve();
				return resolved_preq;
			}

			@Override
			public Component getDisplayName() {
				if (custom_name != null)
					return custom_name;

				return super.getDisplayName();
			}
		};
	}
}
