package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.client.Minecraft
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.ByteArrayTag
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.nbt.ListTag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.CapabilityEnergy
import net.minecraftforge.energy.IEnergyStorage
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.BlockEnergyCounter
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.extractEnergy
import ru.dbotthepony.mc.otm.capability.receiveEnergy
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.menu.MenuEnergyCounter
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import java.lang.ref.WeakReference
import java.util.function.Supplier

data class EnergyCounterPacket(val pos: BlockPos, val thisTick: ImpreciseFraction, val total: ImpreciseFraction, val index: Int, val value: ImpreciseFraction) {
	fun write(buff: FriendlyByteBuf) {
		buff.writeBlockPos(pos)
		thisTick.write(buff)
		total.write(buff)
		buff.writeInt(index)
		value.write(buff)
	}

	fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true

		context.get().enqueueWork {
			val ply = Minecraft.getInstance().player ?: return@enqueueWork
			val level = ply.level ?: return@enqueueWork
			val tile = level.getBlockEntity(pos) as? BlockEntityEnergyCounter ?: return@enqueueWork

			tile.lastTick = thisTick
			tile.passed = total
			tile[index] = value
			tile.historyTick = index
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): EnergyCounterPacket {
			val pos = buff.readBlockPos()
			val thisTick = ImpreciseFraction.read(buff)
			val total = ImpreciseFraction.read(buff)
			val index = buff.readInt()
			val value = ImpreciseFraction.read(buff)
			return EnergyCounterPacket(pos, thisTick, total, index, value)
		}
	}
}

class BlockEntityEnergyCounter(p_155229_: BlockPos, p_155230_: BlockState) : BlockEntityMattery(MBlockEntities.ENERGY_COUNTER, p_155229_, p_155230_) {
	var passed = ImpreciseFraction.ZERO
		internal set

	private val history = Array(10 * 20) { ImpreciseFraction.ZERO }
	var historyTick = 0
		internal set

	fun size() = history.size
	operator fun get(i: Int) = history[i]
	internal operator fun set(i: Int, value: ImpreciseFraction) {
		history[i] = value
	}

	var lastTick = ImpreciseFraction.ZERO
		internal set

	fun getHistory(ticks: Int): Array<ImpreciseFraction> {
		require(!(ticks < 1 || ticks >= history.size)) { "Invalid history length provided" }

		val history = Array(ticks) { ImpreciseFraction.ZERO }

		for (i in 0 until ticks) {
			var index = (historyTick - i) % this.history.size
			if (index < 0) index = this.history.size - 1
			history[i] = this.history[index]
		}

		return history
	}

	fun calcAverage(ticks: Int): ImpreciseFraction {
		return sumHistory(ticks) / ticks
	}

	fun sumHistory(ticks: Int): ImpreciseFraction {
		require(!(ticks < 1 || ticks >= history.size)) { "Invalid history length provided" }

		var value = ImpreciseFraction.ZERO

		for (i in 0 until ticks) {
			var index = (historyTick - i) % history.size
			if (index < 0) index = history.size - 1
			value += history[index]
		}

		return value
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["passed"] = passed.serializeNBT()

		val list = ListTag()
		nbt["history"] = list
		nbt["history_tick"] = historyTick

		for (num in history)
			list.add(num.serializeNBT())
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas(("passed")) {
			passed = ImpreciseFraction.deserializeNBT(it)
		}

		nbt.ifHas(("history_tick"), IntTag::class.java) {
			historyTick = it.asInt
		}

		nbt.ifHas("history", ListTag::class.java) {
			for (i in it.indices) {
				val bytes = it[i] as? ByteArrayTag
				history[i] = if (bytes != null) ImpreciseFraction.deserializeNBT(bytes) else ImpreciseFraction.ZERO
			}
		}
	}

	override fun getDefaultDisplayName(): Component {
		return NAME
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MenuEnergyCounter(containerID, inventory, this)
	}

	private val energyInput = EnergyCounterCap(true)
	private val energyOutput = EnergyCounterCap(false)

	private var inputCapability = LazyOptional.empty<IEnergyStorage>()
	private var outputCapability = LazyOptional.empty<IEnergyStorage>()

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		tickOnceServer(this::checkSurroundings)
	}

	private inner class EnergyCounterCap(val is_input: Boolean) : IMatteryEnergyStorage {
		override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (is_input)
				return ImpreciseFraction.ZERO

			if (inputCapability.isPresent) {
				val it = inputCapability.resolve().get()

				val diff: ImpreciseFraction

				if (it is IMatteryEnergyStorage) {
					diff = it.extractEnergyOuter(howMuch, simulate)
				} else {
					diff = ImpreciseFraction(it.extractEnergy(howMuch, simulate))
				}

				if (!simulate) {
					passed += diff
					history[historyTick] += diff
				}

				return diff
			}

			return ImpreciseFraction.ZERO
		}

		override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (!is_input)
				return ImpreciseFraction.ZERO

			if (outputCapability.isPresent) {
				val it = outputCapability.resolve().get()

				val diff: ImpreciseFraction

				if (it is IMatteryEnergyStorage) {
					diff = it.receiveEnergyOuter(howMuch, simulate)
				} else {
					diff = ImpreciseFraction(it.receiveEnergy(howMuch, simulate))
				}

				if (!simulate) {
					passed += diff
					history[historyTick] += diff
				}

				return diff
			}

			return ImpreciseFraction.ZERO
		}

		override val batteryLevel: ImpreciseFraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.batteryLevel
						}

						return ImpreciseFraction(it.energyStored)
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.batteryLevel
						}

						return ImpreciseFraction(it.energyStored)
					}
				}

				return ImpreciseFraction.ZERO
			}

		override val maxBatteryLevel: ImpreciseFraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.maxBatteryLevel
						}

						return ImpreciseFraction(it.maxEnergyStored)
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.maxBatteryLevel
						}

						return ImpreciseFraction(it.maxEnergyStored)
					}
				}

				return ImpreciseFraction.ZERO
			}

		override val missingPower: ImpreciseFraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.missingPower
						}

						return ImpreciseFraction((it.maxEnergyStored - it.energyStored).coerceAtLeast(0))
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.missingPower
						}

						return ImpreciseFraction((it.maxEnergyStored - it.energyStored).coerceAtLeast(0))
					}
				}

				return ImpreciseFraction.ZERO
			}

		override fun canExtract() = !is_input
		override fun canReceive() = is_input
	}

	private var resolverInput = LazyOptional.of<IMatteryEnergyStorage> { energyInput }
	private var resolverOutput = LazyOptional.of<IMatteryEnergyStorage> { energyOutput }
	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolverInput.invalidate()
		resolverOutput.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolverInput = LazyOptional.of { energyInput }
		resolverOutput = LazyOptional.of { energyOutput }
	}

	@Suppress("deprecation")
	override fun setBlockState(new: BlockState) {
		val old = blockState
		super.setBlockState(new)

		if (new !== old && new.getValue(BlockEnergyCounter.INPUT_DIRECTION) != old.getValue(BlockEnergyCounter.INPUT_DIRECTION)) {
			resolverInput.invalidate()
			resolverOutput.invalidate()
			resolverInput = LazyOptional.of { energyInput }
			resolverOutput = LazyOptional.of { energyOutput }

			if (level != null)
				checkSurroundings(level)
		}
	}

	private fun getAndBind(
		level: Level,
		old: LazyOptional<IEnergyStorage>,
		side: Direction
	): LazyOptional<IEnergyStorage> {
		val ent = level.getBlockEntity(blockPos.offset(side.normal)) ?: return LazyOptional.empty()
		val resolve = ent.getCapability(CapabilityEnergy.ENERGY, side.opposite)

		if (resolve !== old) {
			val weak = WeakReference(this)

			resolve.addListener {
				val get = weak.get()

				if (get?.level != null) {
					get.checkSurroundings(get.level)
				}
			}

			return resolve
		}

		return old
	}

	fun checkSurroundings(level: Level?) {
		if (isRemoved || level !is ServerLevel) return

		inputCapability = getAndBind(
			level,
			inputCapability,
			blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION)
		)

		outputCapability = getAndBind(
			level,
			outputCapability,
			-blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION)
		)
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (side == null || isRemoved)
			return super.getCapability(cap, side)

		if (valid) {
			if (side == blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION)) {
				if (cap === MatteryCapability.ENERGY || cap === CapabilityEnergy.ENERGY) {
					return resolverInput.cast()
				}
			} else if (side == blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION).opposite) {
				if (cap === MatteryCapability.ENERGY || cap === CapabilityEnergy.ENERGY) {
					return resolverOutput.cast()
				}
			}
		}

		return super.getCapability(cap, side)
	}

	private fun distributor(): PacketDistributor.TargetPoint {
		val x = blockPos.x.toDouble()
		val y = blockPos.y.toDouble()
		val z = blockPos.z.toDouble()
		return PacketDistributor.TargetPoint(x, y, z, 32.0, level!!.dimension())
	}

	fun tick() {
		lastTick = history[historyTick]
		val index = historyTick
		val accumulated = history[historyTick]
		historyTick = (historyTick + 1) % history.size
		history[historyTick] = ImpreciseFraction.ZERO

		MatteryNetworking.CHANNEL.send(PacketDistributor.NEAR.with(this::distributor), EnergyCounterPacket(blockPos, lastTick, passed, index, accumulated))
	}

	fun clientTick() {
		passed += lastTick
	}

	companion object {
		private val NAME = TranslatableComponent("block.overdrive_that_matters.energy_counter")
	}
}