package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.items.CapabilityItemHandler
import ru.dbotthepony.mc.otm.block.entity.worker.BlockEntityMatteryWorker
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerJob
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerJobStatus
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerTickContext
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.*
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerFilterOnlyOut
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.matter.baselineComplexityReplicateTicks
import ru.dbotthepony.mc.otm.matter.getMatterValue
import ru.dbotthepony.mc.otm.menu.MenuMatterReplicator
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.set

class BlockEntityMatterReplicator(p_155229_: BlockPos, p_155230_: BlockState) :
	BlockEntityMatteryWorker(MBlockEntities.MATTER_REPLICATOR, p_155229_, p_155230_), IMatterGraphNode {

	override val energy = WorkerEnergyStorage(this, STORAGE, MAX_IO)
	private val node = Graph6Node<IMatterGraphNode>(this)
	private val resolverNode = LazyOptional.of { this }

	override fun getAsMatterNode(): Graph6Node<IMatterGraphNode> {
		return node
	}

	@JvmField
	val matter = MatterHandlerImpl(
		this::setChangedLight,
		MatterDirection.RECEIVE,
		ImpreciseFraction(2)
	)

	// обычные запросы
	@JvmField
	val container = MatteryContainer(this::setChangedLight, 5)
	private val itemHandler = container.handler(MatteryContainerFilterOnlyOut)

	override fun getDefaultDisplayName(): Component {
		return NAME
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MenuMatterReplicator(containerID, inventory, this)
	}

	override fun onJobFinish(job: WorkerJob): WorkerJobStatus {
		if (job.data.getBoolean("as_dust")) {
			val matterValue = moveMatterAsDustIntoContainer(ImpreciseFraction.deserializeNBT(job["matter"]), container, OUTPUT_DUST_MAIN, OUTPUT_DUST_STACKING)

			if (!matterValue.isZero) {
				job["matter"] = matterValue.serializeNBT()
				return WorkerJobStatus(false, 20)
			}

			(node.graph as MatterNetworkGraph?)?.notifyTaskCompletion(MatterTask.deserializeNBT(job["task"])!!)

			return WorkerJobStatus()
		}

		if (!container.fullyAddItem(job.stack, FIRST_ACTUAL_OUTPUT_SLOT .. LAST_ACTUAL_OUTPUT_SLOT)) {
			return WorkerJobStatus(false, 20)
		}

		(node.graph as MatterNetworkGraph?)?.notifyTaskCompletion(MatterTask.deserializeNBT(job["task"])!!)
		return WorkerJobStatus()
	}

	override fun onMatterTaskCreated(task: MatterTask) {
		isIdling = false
	}

	override fun onMatterTaskUpdated(new_state: MatterTask, old_state: MatterTask) {
		isIdling = false
	}

	override fun onPatternAdded(state: PatternState) {
		isIdling = false
	}

	override fun setRemoved() {
		super.setRemoved()
		node.destroy(::MatterNetworkGraph)
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, node)
	}

	override fun computeNextJob(): WorkerJob? {
		val graph = node.graph as MatterNetworkGraph? ?: return null
		val allocation = graph.allocateTask(false) ?: return null
		val stack = allocation.task.stack(1)
		val matter = getMatterValue(stack)

		// ????????
		if (matter.isZero) return null

		val ticks = matter.complexity * baselineComplexityReplicateTicks

		return WorkerJob(stack, ticks, BASE_CONSUMPTION, CompoundTag().also {
			it["matter_per_tick"] = (matter.value / ticks).serializeNBT()
			it["task"] = allocation.task.serializeNBT()
			it["matter"] = matter.value.serializeNBT()

			if (allocation.pattern != null)
				it["pattern"] = allocation.pattern.serializeNBT()

			if ((level?.random?.nextDouble() ?: 1.0) > (allocation.pattern?.research ?: 2.0))
				it["as_dust"] = true
		})
	}

	override fun onWorkTick(context: WorkerTickContext): WorkerJobStatus {
		val drainPerTick = ImpreciseFraction.deserializeNBT(context.job.data["matter_per_tick"])
		val graph = node.graph as MatterNetworkGraph? ?: return WorkerJobStatus(false, 20)

		if (matter.extractMatterInner(drainPerTick, true) < drainPerTick) {
			// в машине недостаточно материи

			if (drainPerTick > matter.maxStoredMatter) {
				// в тик требуется больше материи, чем её может хранить репликатор
				val toExtract = drainPerTick - matter.extractMatterInner(drainPerTick, true)
				val drain = graph.extractMatter(toExtract, true)

				if (drain < toExtract) {
					// недостаточно материи в сети
					return WorkerJobStatus(false, 200)
				}

				// достаточно материи в сети + внутри машины
				matter.extractMatterInner(drainPerTick, false)
				graph.extractMatter(drain, false)
				return WorkerJobStatus()
			} else {
				// в тик требуется меньше материи, чем её может хранить репликатор
				// примем из сети недостающее количество бака материи, или 200 тиков репликации, что меньше
				val toExtract =
					matter.missingMatter.min(drainPerTick.times(DRAIN_MULT))

				val drain = graph.extractMatter(toExtract, true)

				if (drain.isZero) {
					// в сети нет материи
					return WorkerJobStatus(false, 200)
				}

				val received = matter.receiveMatterOuter(drain, false)
				graph.extractMatter(received, false)

				// получили материю, проверяем возможность работы
				if (matter.extractMatterInner(drainPerTick, false) >= drainPerTick) {
					return WorkerJobStatus()
				} else {
					// :(
					return WorkerJobStatus(false, 200)
				}
			}
		}

		// в машине достаточно материи
		matter.extractMatterInner(drainPerTick, false)
		return WorkerJobStatus()
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["container"] = container.serializeNBT()
		nbt["matter"] = matter.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		container.deserializeNBT(nbt["container"])

		nbt.ifHas("matter", CompoundTag::class.java) {
			matter.deserializeNBT(it)
		}
	}

	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		itemHandler.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		itemHandler.revive()
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (valid) {
			if (cap === MatteryCapability.MATTER_NODE) return resolverNode.cast()
			if (cap === CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) return itemHandler.get().cast()
		}

		return super.getCapability(cap, side)
	}

	companion object {
		private val NAME = TranslatableComponent("block.overdrive_that_matters.matter_replicator")
		private val BASE_CONSUMPTION = ImpreciseFraction(400)
		private val DRAIN_MULT = ImpreciseFraction(200)
		private val STORAGE = ImpreciseFraction(200_000)
		private val MAX_IO = ImpreciseFraction(4_000)

		const val FIRST_ACTUAL_OUTPUT_SLOT = 0
		const val LAST_ACTUAL_OUTPUT_SLOT = 2
		const val OUTPUT_DUST_MAIN = 3
		const val OUTPUT_DUST_STACKING = 4
	}
}
