package ru.dbotthepony.mc.otm.block.entity.blackhole

import net.minecraft.client.Minecraft
import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.network.Connection
import net.minecraft.network.protocol.PacketFlow
import net.minecraft.network.protocol.game.ClientboundBlockEntityDataPacket
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.item.ItemEntity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.level.Explosion
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.Blocks
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.levelgen.structure.BoundingBox
import net.minecraft.world.phys.AABB
import net.minecraft.world.phys.Vec3
import ru.dbotthepony.mc.otm.registry.Registry
import ru.dbotthepony.mc.otm.block.BlockBlackHole
import ru.dbotthepony.mc.otm.block.entity.BlockEntityGravitationStabilizer
import ru.dbotthepony.mc.otm.block.entity.blackhole.ExplosionQueue.Companion.queueForLevel
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.plus
import ru.dbotthepony.mc.otm.matter.getMatterValue
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MItems
import ru.dbotthepony.mc.otm.set
import kotlin.math.roundToInt
import kotlin.math.sqrt

class BlockEntityBlackHole(p_155229_: BlockPos, p_155230_: BlockState) : BlockEntity(MBlockEntities.BLACK_HOLE, p_155229_, p_155230_) {
	var mass = BASELINE_MASS
		set(mass) {
			if (mass <= ImpreciseFraction.ZERO) {
				collapse()
				return
			}

			field = mass

			setChanged()

			var massForDiv = mass

			when (if (level?.isClientSide == false) stabilizers.size else stabilizerClientCount) {
				1 -> {massForDiv /= 4}
				2 -> {massForDiv /= 16}
				3 -> {massForDiv /= 64}
				4 -> {massForDiv /= 512}
				5 -> {massForDiv /= 2048}
				6 -> {massForDiv /= 16384}
			}

			gravitationStrength = sqrt(massForDiv.div(BASELINE_MASS).toDouble()).coerceAtMost(20.0).coerceAtLeast(0.2)

			val level = level

			if (level != null && !level.isClientSide && !suppressUpdates)
				level.sendBlockUpdated(blockPos, blockState, blockState, Block.UPDATE_CLIENTS)

			affectedBounds = BoundingBox(
				(-30 * gravitationStrength).toInt(),
				(-30 * gravitationStrength).toInt(),
				(-30 * gravitationStrength).toInt(),
				(30 * gravitationStrength).toInt(),
				(30 * gravitationStrength).toInt(),
				(30 * gravitationStrength).toInt()
			).move(
				blockPos
			)

			affectedBoundsAABB = AABB.of(affectedBounds)
		}

	override fun getRenderBoundingBox(): AABB {
		return AABB(blockPos.offset(-BlockEntityGravitationStabilizer.RANGE, -BlockEntityGravitationStabilizer.RANGE, -BlockEntityGravitationStabilizer.RANGE), blockPos.offset(BlockEntityGravitationStabilizer.RANGE, BlockEntityGravitationStabilizer.RANGE, BlockEntityGravitationStabilizer.RANGE))
	}

	var gravitationStrength = 1.0
		private set

	private var suppressUpdates = true
	var spin_direction = false

	private var sleepTicks = 4

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		sleepTicks = 4
	}

	private val stabilizers = ArrayList<BlockEntityGravitationStabilizer>()
	private var stabilizerClientCount = 0

	fun stabilizerAttached(stabilizer: BlockEntityGravitationStabilizer) {
		if (stabilizers.contains(stabilizer))
			return

		stabilizers.add(stabilizer)
		mass = mass
	}

	fun stabilizerDetached(stabilizer: BlockEntityGravitationStabilizer) {
		if (stabilizers.remove(stabilizer)) {
			mass = mass
		}
	}

	fun collapse() {
		val level = level as? ServerLevel ?: return

		level.setBlock(blockPos, Blocks.AIR.defaultBlockState(), Block.UPDATE_ALL)

		if (gravitationStrength > 0.25) {
			val x0 = blockPos.x + 0.5
			val y0 = blockPos.y + 0.5
			val z0 = blockPos.z + 0.5
			val queue = queueForLevel(level)
			var radius = 0

			while (radius < Math.ceil(gravitationStrength * 4)) {
				queue.explodeRing(
					x0,
					y0,
					z0,
					radius.toDouble(), Math.min(20.0, Math.max(1.0, (gravitationStrength * 4 - radius) * 20)).toFloat()
				)

				radius++
			}
		} else {
			level.explode(
				null,
				Registry.DAMAGE_HAWKING_RADIATION,
				null,
				blockPos.x + 0.5,
				blockPos.y + 0.5,
				blockPos.z + 0.5,
				gravitationStrength.toFloat() * 60,
				false,
				Explosion.BlockInteraction.DESTROY
			)
		}
	}

	// shared functions
	fun writeBlackHoleData(tag: CompoundTag): CompoundTag {
		tag["mass"] = mass.serializeNBT()
		tag["spin_direction"] = spin_direction
		return tag
	}

	fun readBlackHoleData(tag: CompoundTag) {
		tag["mass"]?.let {
			mass = ImpreciseFraction.deserializeNBT(it)
		}

		spin_direction = tag.getBoolean("spin_direction")
	}

	// disk io
	public override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		writeBlackHoleData(nbt)
	}

	override fun load(tag: CompoundTag) {
		super.load(tag)
		readBlackHoleData(tag)
	}

	// received either by game engine (from getUpdateTag on ClientLevelPacket)
	// or by onDataPacket
	override fun handleUpdateTag(tag: CompoundTag) {
		stabilizerClientCount = (tag["stabilizers"] as IntTag).asInt
		super.handleUpdateTag(tag)
		readBlackHoleData(tag)
	}

	// called by game engine for ClientLevelPacket
	override fun getUpdateTag(): CompoundTag {
		return writeBlackHoleData(super.getUpdateTag()).also {
			it["stabilizers"] = stabilizers.size
		}
	}

	// called by game engine on block updates
	override fun getUpdatePacket(): ClientboundBlockEntityDataPacket? {
		return ClientboundBlockEntityDataPacket.create(this)
	}

	// called by game engine by forge patches when ClientboundBlockEntityDataPacket is received
	override fun onDataPacket(net: Connection, pkt: ClientboundBlockEntityDataPacket) {
		super.onDataPacket(net, pkt)

		if (net.receiving == PacketFlow.CLIENTBOUND) {
			handleUpdateTag(pkt.tag!!)
		}
	}

	var affectedBounds = BoundingBox(0, 0, 0, 1, 1, 1)
		private set
	var affectedBoundsAABB = AABB.of(affectedBounds)
		private set

	private fun setDeltaMovement(living: Entity, center: Vec3, distance: Double, weaker: Boolean) {
		//final double mult = Math.min(2, (30 * this.gravitation_strength) / Math.max(1, Math.pow(distance, 2)));
		// Сила притяжения
		val mult = Math.pow((1 - distance / (30 * gravitationStrength)).coerceAtLeast(0.0), 2.0) * gravitationStrength / 8

		// Притяжение к ядру
		val delta = living.position().vectorTo(center).normalize()

		if (distance < gravitationStrength) {
			living.deltaMovement = living.deltaMovement.add(delta.multiply(mult, mult, mult))
		} else {
			// Закручивание
			val rotate = if (spin_direction) delta.yRot((Math.PI / 2).toFloat()) else delta.yRot((-Math.PI / 2).toFloat())

			if (weaker)
				living.deltaMovement = living.deltaMovement
					.add(rotate.multiply(mult * 0.2f, mult * 0.2f, mult * 0.2f))
					.add(delta.multiply(mult * 0.33f, mult * 0.33f, mult * 0.33f))
			else
				living.deltaMovement = living.deltaMovement
					.add(rotate.multiply(mult * 0.2f, mult * 0.2f, mult * 0.2f))
					.add(delta.multiply(mult, mult, mult))
		}
	}

	fun clientTick() {
		sleepTicks--
		if (sleepTicks > 0) return
		val ply = Minecraft.getInstance().player!!
		val center = Vec3.atCenterOf(blockPos)

		if (!ply.abilities.mayfly) {
			val distance = ply.position().distanceTo(center)
			setDeltaMovement(ply, center, distance, true)
		}

		for (item in level!!.getEntitiesOfClass(ItemEntity::class.java, affectedBoundsAABB)) {
			val distance = item.position().distanceTo(center)
			setDeltaMovement(item, center, distance, false)
		}
	}

	fun tick() {
		sleepTicks--
		if (sleepTicks > 0) return
		val level = level as? ServerLevel ?: return
		suppressUpdates = false

		val center = Vec3.atCenterOf(blockPos)

		for (living in level.getEntitiesOfClass(LivingEntity::class.java, affectedBoundsAABB)) {
			val distance = living.position().distanceTo(center)

			if (living !is Player || !living.abilities.mayfly) {
				setDeltaMovement(living, center, distance, living !is Player)
			}

			if (distance < gravitationStrength + 1) {
				living.hurt(Registry.DAMAGE_EVENT_HORIZON, (gravitationStrength / distance).toFloat())
			}
		}

		for (item in level.getEntitiesOfClass(ItemEntity::class.java, affectedBoundsAABB)) {
			val distance = item.position().distanceTo(center)
			setDeltaMovement(item, center, distance, false)

			if (distance < gravitationStrength + 1) {
				if (item.hurt(Registry.DAMAGE_EVENT_HORIZON, (gravitationStrength / distance).toFloat()) && item.isRemoved) {
					if (item.item.item === MItems.GRAVITATIONAL_DISRUPTOR) {
						collapse()
					} else {
						val mass = getMatterValue(item.item)

						if (!mass.isZero)
							this.mass += mass.value * item.item.count
					}
				}
			}
		}

		// шанс 1% что черная дыра потеряет 0.1 MtU каждую секунду * силу гравитации дыры ^ -1
		if (level.random.nextDouble() < 0.01 * 0.05 * (1 / gravitationStrength)) {
			this.mass += HAWKING_MASS_LOSE_STEP
		}

		if (gravitationStrength > 0.4f) {
			val sphere = getSphericalShape((gravitationStrength * 6.0).roundToInt())

			if (sphere.size != lastSphereSizeOuter) {
				lastSphereSizeOuter = sphere.size
				sphereIndexOuter = 0
			}

			var iterations = 0

			while (iterations < ITERATIONS) {
				iterations++

				val pos = sphere[sphereIndexOuter] + blockPos
				sphereIndexOuter = (sphereIndexOuter + 1) % lastSphereSizeOuter

				if (sphereIndexOuter == 0 && lastSphereSizeOuter <= ITERATIONS)
					break

				val getBlock = level.getBlockState(pos)

				if (!getBlock.isAir && getBlock.block !is BlockBlackHole) {
					val speed = getBlock.getDestroySpeed(level, pos)

					val eResist = try {
						getBlock.getExplosionResistance(level, pos, null)
					} catch (err: Throwable) {
						getBlock.block.getExplosionResistance()
						// Потому что возможно какой-либо мод не ожидает что Explosion == null
						// особенно учитывая что интерфейс IForgeBlock не имеет @ParamsAreNonnullByDefault
						// и аргумент не помечен как @Nullable
						// тем самым имеет тип Explosion! который указывается как Explosion? .. Explosion!!
					}

					var strengthLinear = sqrt(blockPos.distSqr(pos))

					if (strengthLinear < gravitationStrength * 1.1) {
						strengthLinear /= 4.0
					}

					if (speed >= 0f && (speed * 2f).coerceAtLeast(eResist / 3f) < gravitationStrength * (8.0 / strengthLinear)) {
						Block.dropResources(getBlock, level, pos, level.getBlockEntity(pos))
						getBlock.block.destroy(level, pos, getBlock)
						level.setBlock(pos, getBlock.fluidState.createLegacyBlock(), Block.UPDATE_ALL)
					}
				}
			}
		}
	}

	private var lastSphereSizeOuter = 0
	private var sphereIndexOuter = 0

	init {
		mass = mass
	}

	companion object {
		const val ITERATIONS = 30_000
		val BASELINE_MASS = ImpreciseFraction(1_000)
		val HAWKING_MASS_LOSE_STEP = ImpreciseFraction("-0.1")

		private val blockShapeCache = HashMap<Int, Array<BlockPos>>()

		private fun getSphericalShape(size: Int): Array<BlockPos> {
			return blockShapeCache.computeIfAbsent(size) {
				val result = ArrayList<BlockPos>((it * it * it * Math.PI * (4.0 / 3.0)).toInt() + 16)

				for (x in -it .. it) {
					for (y in -it .. it) {
						for (z in -it .. it) {
							val dist = sqrt(x * x.toDouble() + y * y.toDouble() + z * z.toDouble())

							if (dist <= size && dist != 0.0) {
								result.add(BlockPos(x, y, z))
							}
						}
					}
				}

				val arr = result.toTypedArray()

				arr.sortWith { a, b ->
					val xa = a.x
					val ya = a.y
					val za = a.z
					val distA = sqrt(xa * xa.toDouble() + ya * ya.toDouble() + za * za.toDouble())

					val xb = b.x
					val yb = b.y
					val zb = b.z
					val distB = sqrt(xb * xb.toDouble() + yb * yb.toDouble() + zb * zb.toDouble())

					if (distA == distB) {
						return@sortWith 0
					} else if (distA > distB) {
						return@sortWith 1
					} else {
						return@sortWith -1
					}
				}

				return@computeIfAbsent arr
			}
		}
	}
}