package ru.dbotthepony.mc.otm.capability

import net.minecraft.nbt.CompoundTag
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.set

private enum class MachineType {
	WORKER, GENERATOR, CAPACITOR
}

sealed class EnergyStorageImpl constructor(
	protected val listener: () -> Unit,
	private val type: MachineType,
	override var maxBatteryLevel: ImpreciseFraction,
	protected var maxInput: ImpreciseFraction,
	protected var maxOutput: ImpreciseFraction
) : IMatteryEnergyStorage, INBTSerializable<CompoundTag> {
	override var batteryLevel = ImpreciseFraction.ZERO
		protected set

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (type == MachineType.WORKER)
			return ImpreciseFraction.ZERO

		return extractEnergyInner(howMuch, simulate)
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		val new = batteryLevel.minus(howMuch.min(maxOutput)).moreThanZero()
		val diff = batteryLevel.minus(new)

		if (!simulate && batteryLevel != new) {
			batteryLevel = new
			listener()
		}

		return diff
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		if (type == MachineType.GENERATOR)
			return ImpreciseFraction.ZERO

		return receiveEnergyInner(howMuch, simulate)
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		val new = batteryLevel.plus(howMuch.min(maxInput)).min(maxBatteryLevel)
		val diff = new.minus(batteryLevel)

		if (!simulate && batteryLevel != new) {
			batteryLevel = new
			listener()
		}

		return diff
	}

	override fun canExtract(): Boolean {
		return type != MachineType.WORKER
	}

	override fun canReceive(): Boolean {
		return type != MachineType.GENERATOR
	}

	override fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()
		// TODO: А это вообще надо?
		tag["energy_stored"] = batteryLevel.serializeNBT()

		// tag["energy_stored_max"] = maxBatteryLevel.serializeNBT()
		// tag["max_input"] = maxInput.serializeNBT()
		// tag["max_output"] = maxOutput.serializeNBT()

		return tag
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		nbt.ifHas("energy_stored") { batteryLevel = ImpreciseFraction.deserializeNBT(it) }
		// nbt.ifHas("energy_stored_max") { maxBatteryLevel = deserializeNBT(it) }
		// nbt.ifHas("max_input") { maxInput = deserializeNBT(it) }
		// nbt.ifHas("max_output") { maxOutput = deserializeNBT(it) }
	}

	var resolver: LazyOptional<IMatteryEnergyStorage> = LazyOptional.of { this }
		private set

	fun invalidate() {
		resolver.invalidate()
	}

	fun revive() {
		resolver = LazyOptional.of { this }
	}

	companion object {
		val DEFAULT_MAX_IO = ImpreciseFraction(200)
		val DEFAULT_MAX_CAPACITY = ImpreciseFraction(60_000)
	}
}

open class WorkerEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction = maxInput
) : EnergyStorageImpl(listener, MachineType.WORKER, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)
}

open class GeneratorEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction = maxInput
) : EnergyStorageImpl(listener, MachineType.GENERATOR, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)
}

open class CapacitorEnergyStorage(
	listener: () -> Unit,
	maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
	maxInput: ImpreciseFraction = DEFAULT_MAX_IO,
	maxOutput: ImpreciseFraction = maxInput
) : EnergyStorageImpl(listener, MachineType.CAPACITOR, maxBatteryLevel, maxInput, maxOutput) {
	constructor(
		listener: BlockEntity,
		maxBatteryLevel: ImpreciseFraction = DEFAULT_MAX_CAPACITY,
		maxInput: ImpreciseFraction = DEFAULT_MAX_IO,
		maxOutput: ImpreciseFraction = maxInput) : this({listener.setChanged()}, maxBatteryLevel, maxInput, maxOutput)
}
