package ru.dbotthepony.mc.otm.capability.drive

import ru.dbotthepony.mc.otm.core.Fraction.Companion.deserializeNBT
import ru.dbotthepony.mc.otm.storage.IStorageStack
import kotlin.jvm.JvmOverloads
import java.util.HashMap
import ru.dbotthepony.mc.otm.storage.IStorageTuple
import java.util.UUID
import ru.dbotthepony.mc.otm.storage.StorageObjectTuple
import ru.dbotthepony.mc.otm.storage.IStorageListener
import ru.dbotthepony.mc.otm.storage.StorageTuple
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.set
import java.util.ArrayList
import java.util.HashSet


abstract class AbstractMatteryDrive<T : IStorageStack> @JvmOverloads constructor(
	override var driveCapacity: ImpreciseFraction,
	override val uuid: UUID = UUID.randomUUID(),
	var maxDifferentStacks: Int = 0xFFFF
) : IMatteryDrive<T> {

	@JvmField
	protected val storedStacks = HashMap<Any, MutableList<IStorageTuple<T>>>()

	@JvmField
	protected val storedByID = HashMap<UUID, IStorageTuple<T>>()
	abstract fun identity(): StorageObjectTuple<T>
	override var isDirty = false
		set(value) {
			if (value != field && value && DrivePool.isLegalAccess()) {
				DrivePool.markDirty(uuid)
			}

			field = value
		}

	var storedDifferentStacks = 0

	override var storedCount = ImpreciseFraction.ZERO
		protected set

	override fun insertStack(stack: T, simulate: Boolean): T {
		val maxInsert = driveCapacity.minus(storedCount).min(stack.count)
		if (maxInsert <= ImpreciseFraction.ZERO) return stack

		val listing = storedStacks.computeIfAbsent(stack.partitionKey()) { ArrayList() }

		for (state in listing) {
			if (state.stack().sameItem(stack)) {
				if (!simulate) {
					state.stack().grow(maxInsert)
					storedCount += maxInsert

					for (listener in listeners) {
						listener.changeObject(state.id(), state.stack().count)
					}

					isDirty = true
				}

				val copy = stack.copy() as T
				copy.shrink(maxInsert)
				return copy
			}
		}

		if (storedDifferentStacks >= maxDifferentStacks) {
			return stack
		}

		if (!simulate) {
			storedDifferentStacks++
			storedCount = storedCount.plus(maxInsert)

			val copy = stack.copy() as T
			copy.count = maxInsert

			val state = StorageTuple(UUID.randomUUID(), copy)
			listing.add(state)
			storedByID[state.id()] = state

			for (listener in listeners) {
				listener.addObject(state.stack(), state.id(), this)
			}

			isDirty = true
		}

		val copy = stack.copy() as T
		copy.shrink(maxInsert)
		return copy
	}

	override fun extractStack(id: UUID, amount: ImpreciseFraction, simulate: Boolean): T {
		var amount = amount
		val get = storedByID[id] ?: return identity().empty()

		if (amount <= ImpreciseFraction.ZERO)
			amount = get.stack.getMaxStackSize().orElse(get.stack.count)

		amount = amount.min(get.stack.count)

		if (amount <= ImpreciseFraction.ZERO)
			return identity().empty()

		val copy = get.stack.copy() as T
		copy.count = amount

		if (!simulate) {
			if (amount.compareTo(get.stack.count) == 0) {
				val listing = storedStacks[get.stack.partitionKey()]!!
				listing.remove(get)
				storedDifferentStacks--

				for (listener in listeners) {
					listener.removeObject(get.id())
				}

				if (listing.size == 0) {
					storedStacks.remove(get.stack.partitionKey())
				}
			}

			storedCount = storedCount.minus(amount)
			get.stack.shrink(amount)

			if (!get.stack.count.isZero) {
				for (listener in listeners) {
					listener.changeObject(get.id(), get.stack.count)
				}
			}

			isDirty = true
		}

		return copy
	}

	protected abstract fun serializeStack(item: IStorageTuple<T>): CompoundTag?
	protected abstract fun deserializeStack(tag: CompoundTag): T?

	override fun serializeNBT(): CompoundTag {
		val compound = CompoundTag()

		compound["capacity"] = driveCapacity.serializeNBT()
		compound["max_different_stacks"] = maxDifferentStacks

		val list = ListTag()
		compound["items"] = list

		for (value in storedStacks.values) {
			for (stack in value) {
				val serialized = serializeStack(stack)

				if (serialized != null) {
					serialized["id"] = longArrayOf(stack.id().mostSignificantBits, stack.id().leastSignificantBits)
					list.add(serialized)
				}
			}
		}
		return compound
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		storedStacks.clear()
		storedByID.clear()
		storedCount = ImpreciseFraction.ZERO
		storedDifferentStacks = 0

		nbt.ifHas("capacity") {
			driveCapacity = ImpreciseFraction.deserializeNBT(it)
		}

		maxDifferentStacks = nbt.getInt("max_different_stacks")

		for (entry in nbt.getList("items", Tag.TAG_COMPOUND.toInt())) {
			if (entry is CompoundTag) {
				val stack = deserializeStack(entry)

				if (stack != null) {
					storedCount = storedCount.plus(stack.count)
					storedDifferentStacks++

					val id = entry.getLongArray("id")
					val tuple = StorageTuple(if (id.size == 2) UUID(id[0], id[1]) else UUID.randomUUID(), stack)

					storedStacks.computeIfAbsent(stack.partitionKey()) { ArrayList() }.add(tuple)
					storedByID[tuple.id] = tuple
				}
			}
		}
	}

	override fun storageIdentity(): Class<T> {
		return identity().identity()
	}

	override fun getStack(id: UUID): T {
		return storedByID[id]?.stack ?: identity().empty()
	}

	override fun getStacks(): List<IStorageTuple<T>> {
		var amount = 0

		for (listing in storedStacks.values)
			amount += listing.size

		val list = ArrayList<IStorageTuple<T>>(amount)

		for (listing in storedStacks.values) {
			list.addAll(listing)
		}

		return list
	}

	@JvmField
	protected val listeners = HashSet<IStorageListener<T>>()

	override fun addListener(listener: IStorageListener<T>): Boolean {
		return listeners.add(listener)
	}

	override fun removeListener(listener: IStorageListener<T>): Boolean {
		return listeners.remove(listener)
	}
}