package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.Minecraft
import net.minecraft.client.gui.Font
import net.minecraft.client.gui.components.events.GuiEventListener
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.client.render.RenderHelper
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import kotlin.math.max

@JvmRecord
data class ScreenPos(val x: Float, val y: Float)

@JvmRecord
data class DockProperty @JvmOverloads constructor(val left: Float = 0f, val top: Float = 0f, val right: Float = 0f, val bottom: Float = 0f)

enum class Dock {
	NONE, LEFT, RIGHT, TOP, BOTTOM, FILL
}

open class EditablePanel @JvmOverloads constructor(
	@JvmField val screen: MatteryScreen<*>,
	parent: EditablePanel?,

	// относительно родителя
	x: Float = 0f,
	y: Float = 0f,

	width: Float = 10f,
	height: Float = 10f,
) : GuiEventListener {
	var parent: EditablePanel? = null
		set(value) {
			if (field === value)
				return

			field?.onUnParent(this)
			field = value
			needsInvalidation = true
			value?.onParent(this)
		}

	var x: Float = 0f
		set(value) {
			if (field == value) return

			val old = field
			field = value
			needsInvalidation = true
			xPosUpdated(value, old)
		}

	var y: Float = 0f
		set(value) {
			if (field == value) return

			val old = field
			field = value
			needsInvalidation = true
			yPosUpdated(value, old)
		}

	var width: Float = 10f
		set(value) {
			if (field == value) return

			assert(value >= 0f) {"Invalid width $value"}

			val old = field
			field = value
			needsInvalidation = true
			widthUpdated(value, old)
		}

	var height: Float = 10f
		set(value) {
			if (field == value) return

			assert(value >= 0f) {"Invalid height $value"}

			val old = field
			field = value
			needsInvalidation = true
			heightUpdated(value, old)
		}

	@JvmField
	protected val children = ArrayList<EditablePanel>()
	private var needsInvalidation = true

	protected var visibleAsChildren: Boolean = true
		private set

	var visible: Boolean = true
		set(value) {
			if (field != value) {
				val old = isVisible()
				field = value
				val new = isVisible()

				if (old != new) {
					visibilityChanges(new, old)
					parent?.needsInvalidation = true
					needsInvalidation = true

					if (new) {
						killFocus()
					}
				}

				updateVisible()
			}
		}

	// Позволяет смещать потомков на эту координату
	var xOffset = 0f
	var yOffset = 0f

	// Обрезать отрисовку потомков?
	var scissor = false
	var dock = Dock.NONE
		set(value) {
			field = value
			parent?.needsInvalidation = true
		}
	var dockMargin = DockProperty()
		set(value) {
			if (field != value) {
				parent?.needsInvalidation = true
			}

			field = value
		}
	var dockPadding = DockProperty()
		set(value) {
			if (field != value) {
				parent?.needsInvalidation = true
			}

			field = value
		}

	var acceptMouseInput = true
	var acceptKeyboardInput = true
	var ignoreMouseEventBoundaries = false

	@JvmOverloads
	fun setDockMargin(left: Float = 0f, top: Float = 0f, right: Float = 0f, bottom: Float = 0f) {
		dockMargin = DockProperty(left, top, right, bottom)
	}

	@JvmOverloads
	fun setDockPadding(left: Float = 0f, top: Float = 0f, right: Float = 0f, bottom: Float = 0f) {
		dockPadding = DockProperty(left, top, right, bottom)
	}

	// абсолютное начало координат этой панели и потомков
	// негативно если потомок выходит за границы
	var boundingX = 0f
		private set
	var boundingY = 0f
		private set
	var boundingWidth = 0f
		private set
	var boundingHeight = 0f
		private set

	// Абсолютная координата, обновляется на отрисовке
	var absoluteX = 0f
		private set
	var absoluteY = 0f
		private set

	var isHovered = false
		private set

	fun unsetHovered() {
		isHovered = false
	}

	var isFocused = false
		set(value) {
			if (field != value) {
				val old = field
				field = value

				onFocusChanged(old, value)
				findAbsoluteRoot().updateFocus()
			}
		}

	var autoKillFocus = false
		protected set

	private var focusedAsParent = false

	var accumulatedDepth = 0f

	fun accumulate3DHeight(value: Float) {
		accumulatedDepth += value
	}

	fun font(): Font {
		return screen.font
	}

	fun invalidateLayout() {
		needsInvalidation = true
	}

	fun most3DHeight(): Float {
		var depth = accumulatedDepth

		for (child in children) {
			depth = max(depth, child.most3DHeight())
		}

		return depth
	}

	init {
		this.x = x
		this.y = y
		this.width = width
		this.height = height
		this.parent = parent
	}

	private fun onParent(child: EditablePanel) {
		if (children.contains(child)) throw IllegalStateException("Already containing this child")
		children.add(child)
		needsInvalidation = true
		updateVisible()
	}

	private fun onUnParent(child: EditablePanel) {
		if (!children.contains(child)) throw IllegalStateException("Already not containing this child")
		children.remove(child)
		needsInvalidation = true
		updateVisible()
	}

	protected open fun innerRender(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float) {}
	protected open fun innerRenderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		return false
	}

	fun isVisible(): Boolean {
		return visible && visibleAsChildren
	}

	fun render(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Float {
		if (!isVisible()) {
			return 0f
		}

		var depth = 0f

		if (needsInvalidation) {
			needsInvalidation = false
			performLayout()
		}

		val parent = this.parent

		if (parent == null) {
			absoluteX = x
			absoluteY = y
		} else {
			isHovered = parent.isHovered &&
					mouse_x >= absoluteX &&
					mouse_x < absoluteX + width &&
					mouse_y >= absoluteY &&
					mouse_y < absoluteY + height
		}

		val scissor = this.scissor

		if (scissor) {
			val scale = Minecraft.getInstance().window.guiScale

			RenderHelper.pushScissorRect(
				(absoluteX * scale).toInt(),
				(absoluteY * scale).toInt(),
				(width * scale).toInt() + 1,
				(height * scale).toInt() + 1,
			)
		}

		stack.pushPose()
		stack.translate(absoluteX.toDouble(), absoluteY.toDouble(), accumulatedDepth.toDouble())
		innerRender(stack, mouse_x, mouse_y, flag)
		stack.popPose()

		for (child in children) {
			if (child.isVisible()) {
				child.accumulatedDepth = accumulatedDepth + 1
				child.absoluteX = absoluteX + child.x + xOffset
				child.absoluteY = absoluteY + child.y + yOffset

				RenderSystem.setShaderColor(1f, 1f, 1f, 1f)
				depth = max(depth, child.render(stack, mouse_x, mouse_y, flag))
			}
		}

		if (scissor) {
			RenderHelper.popScissorRect()
		}

		return depth
	}

	open fun tickHover(mouse_x: Float, mouse_y: Float): Boolean {
		isHovered =
				mouse_x >= absoluteX &&
				mouse_x <= absoluteX + width &&
				mouse_y >= absoluteY &&
				mouse_y <= absoluteY + height

		return isHovered
	}

	fun renderTooltips(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float): Boolean {
		if (!isVisible()) {
			return false
		}

		if (innerRenderTooltips(stack, mouse_x, mouse_y, flag)) {
			return true
		}

		for (child in children) {
			if (child.isVisible() && child.renderTooltips(stack, mouse_x, mouse_y, flag)) {
				return true
			}
		}

		return false
	}

	@JvmOverloads
	fun localToScreen(_x: Float = 0f, _y: Float = 0f): ScreenPos {
		var parent = this.parent

		var x = this.x + _x
		var y = this.y + _y

		while (parent != null) {
			x += parent.x + parent.xOffset
			y += parent.y + parent.yOffset
			parent = parent.parent
		}

		return ScreenPos(x, y)
	}

	fun screenToLocal(_x: Float = 0f, _y: Float = 0f): ScreenPos {
		val pos = localToScreen()
		return ScreenPos(_x - pos.x, _y - pos.y)
	}

	fun screenToLocal(_x: Double = 0.0, _y: Double = 0.0): ScreenPos {
		val pos = localToScreen()
		return ScreenPos((_x - pos.x).toFloat(), (_y - pos.y).toFloat())
	}

	fun findRoot(): EditablePanel? {
		var parent = this.parent

		if (parent == null) {
			return null
		}

		while (true) {
			val parent2 = parent!!.parent

			if (parent2 != null) {
				parent = parent2
			} else {
				return parent
			}
		}
	}

	fun findAbsoluteRoot(): EditablePanel {
		return findRoot() ?: this
	}

	protected open fun performLayout() {
		var left = dockPadding.left
		var right = dockPadding.right
		var top = dockPadding.top
		var bottom = dockPadding.bottom

		for (child in children) {
			if (child.isVisible()) {
				when (child.dock) {
					Dock.NONE -> {}
					Dock.FILL -> {}

					Dock.LEFT -> {
						child.setPos(left + child.dockMargin.left, top + child.dockMargin.top)
						left += child.width + child.dockMargin.left + child.dockMargin.right
						child.height = (
							Math.max(
								1f,
								height - top - bottom - child.dockMargin.top - child.dockMargin.bottom
							)
						)
					}

					Dock.RIGHT -> {
						child.setPos(
							width - right - child.width - child.dockMargin.right,
							top + child.dockMargin.top
						)

						right += child.width + child.dockMargin.left + child.dockMargin.right
						child.height = (
							Math.max(
								1f,
								height - top - bottom - child.dockMargin.top - child.dockMargin.bottom
							)
						)
					}

					Dock.TOP -> {
						child.setPos(left + child.dockMargin.left, top + child.dockMargin.top)
						top += child.height + child.dockMargin.top + child.dockMargin.bottom

						child.width = (
							Math.max(
								1f,
								width - left - right - child.dockMargin.left - child.dockMargin.right
							)
						)
					}

					Dock.BOTTOM -> {
						child.setPos(
							left + child.dockMargin.left,
							height - bottom - child.height - child.dockMargin.bottom
						)

						bottom += child.height + child.dockMargin.top + child.dockMargin.bottom

						child.width = (
							Math.max(
								1f,
								width - left - right - child.dockMargin.left - child.dockMargin.right
							)
						)
					}
				}
			}
		}

		for (child in children) {
			if (child.isVisible() && child.dock == Dock.FILL) {
				val w = width - child.dockMargin.left - child.dockMargin.right - right - left
				val h = height - child.dockMargin.bottom - child.dockMargin.top - bottom - top

				if (w < 1f || h < 1f) {
					LOGGER.error("Unable to fit $child inside $this (FILL returned dimensions of $w $h)")
				}

				child.setDimensions(
					left + child.dockMargin.left,
					top + child.dockMargin.top,
					Math.max(1f, w),
					Math.max(1f, h)
				)
			}
		}

		boundingX = 0f
		boundingY = 0f
		boundingWidth = 0f
		boundingHeight = 0f

		for (child in children) {
			if (child.isVisible()) {
				boundingX = boundingX.coerceAtMost(child.x)
				boundingY = boundingY.coerceAtMost(child.y)
				boundingWidth = boundingWidth.coerceAtLeast(child.x + child.width)
				boundingHeight = boundingHeight.coerceAtLeast(child.y + child.height)
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()!!
	}

	protected open fun visibilityChanges(new: Boolean, old: Boolean) {

	}

	private fun updateVisible() {
		for (child in children) {
			val old = child.isVisible()
			child.visibleAsChildren = isVisible()
			val new = child.isVisible()

			if (old != new) {
				child.visibilityChanges(new, old)
			}

			child.updateVisible()
		}
	}

	private fun killFocusInternal() {
		for (child in children) {
			child.killFocusInternal()
		}

		if (isFocused) {
			isFocused = false
		}
	}

	fun killFocus() {
		if (isEverFocused()) {
			killFocusInternal()
			findAbsoluteRoot().updateFocus()
		}
	}

	fun findHierarchicalFocus(): EditablePanel? {
		if (isFocused) {
			return this
		}

		for (child in children) {
			val focus = child.findHierarchicalFocus()

			if (focus != null) {
				return focus
			}
		}

		return null
	}

	fun hasHierarchicalFocus(): Boolean {
		return findHierarchicalFocus() != null
	}

	protected open fun onHierarchicalFocusChanged(new: Boolean, old: Boolean) {

	}

	protected open fun onFocusChanged(new: Boolean, old: Boolean) {

	}

	private fun updateFocus() {
		val old = focusedAsParent
		focusedAsParent = hasHierarchicalFocus()

		if (focusedAsParent != old) {
			onHierarchicalFocusChanged(focusedAsParent, old)
		}

		for (child in children) {
			child.updateFocus()
		}
	}

	fun isEverFocused(): Boolean {
		return isFocused || focusedAsParent
	}

	fun requestFocus() {
		if (isFocused) {
			return
		}

		if (focusedAsParent) {
			var child = findHierarchicalFocus()

			while (child != null) {
				child.isFocused = false
				child = findHierarchicalFocus()
			}
		}

		isFocused = true
	}

	protected open fun xPosUpdated(new: Float, old: Float) {}
	protected open fun yPosUpdated(new: Float, old: Float) {}
	protected open fun widthUpdated(new: Float, old: Float) {}
	protected open fun heightUpdated(new: Float, old: Float) {}

	fun setPos(x: Float, y: Float) {
		this.x = x
		this.y = y
	}

	fun setPos(x: Double, y: Double) {
		this.x = x.toFloat()
		this.y = y.toFloat()
	}

	fun setSize(w: Float, h: Float) {
		this.width = w
		this.height = h
	}

	fun setSize(w: Double, h: Double) {
		this.width = w.toFloat()
		this.height = h.toFloat()
	}

	fun setDimensions(x: Float, y: Float, w: Float, h: Float) {
		setPos(x, y)
		setSize(w, h)
	}

	fun setDimensions(x: Double, y: Double, w: Double, h: Double) {
		setPos(x, y)
		setSize(w, h)
	}

	fun fetchChildren(): List<EditablePanel> {
		return java.util.List.copyOf(children)
	}

	fun getUndockedChildren(): List<EditablePanel> {
		var count = 0

		for (child in children) {
			if (child.dock == Dock.NONE) {
				count++
			}
		}

		val list = ArrayList<EditablePanel>(count)

		for (child in children) {
			if (child.dock == Dock.NONE) {
				list.add(child)
			}
		}

		return list
	}

	operator fun get(int: Int): EditablePanel? {
		if (int < 0 || int >= children.size) return null
		return children[int]
	}

	fun isIgnoringMouseEventBoundaries(): Boolean {
		if (ignoreMouseEventBoundaries) {
			return true
		}

		for (child in children) {
			if (child.isIgnoringMouseEventBoundaries()) {
				return true
			}
		}

		return false
	}

	override fun mouseMoved(p_94758_: Double, p_94759_: Double) {

	}

	protected open fun mouseClickedInner(mouse_x: Double, mouse_y: Double, mouse_click_type: Int): Boolean {
		return true
	}

	fun withinBounds(mouse_x: Double, mouse_y: Double): Boolean {
		val pos: ScreenPos = localToScreen()
		return pos.x <= mouse_x && pos.x + width > mouse_x && pos.y <= mouse_y && pos.y + height > mouse_y
	}

	fun withinExtendedBounds(mouse_x: Double, mouse_y: Double): Boolean {
		val pos: ScreenPos = localToScreen(boundingX, boundingY)
		return pos.x <= mouse_x && pos.x + boundingWidth > mouse_x && pos.y <= mouse_y && pos.y + boundingHeight > mouse_y
	}

	final override fun mouseClicked(mouse_x: Double, mouse_y: Double, flag: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (ignoreMouseEventBoundaries) return mouseClickedInner(mouse_x, mouse_y, flag)

		for (child in children) {
			if (child.mouseClickedChecked(mouse_x, mouse_y, flag)) {
				return true
			} else if (child.autoKillFocus) {
				child.killFocus()
			}
		}

		return mouseClickedInner(mouse_x, mouse_y, flag)
	}

	fun mouseClickedChecked(mouse_x: Double, mouse_y: Double, flag: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isIgnoringMouseEventBoundaries() || withinBounds(mouse_x, mouse_y)) {
			if (acceptMouseInput && parent == null) screen.popup(this)
			return mouseClicked(mouse_x, mouse_y, flag)
		} else if (withinExtendedBounds(mouse_x, mouse_y)) {
			for (child in children) {
				if (child.mouseClickedChecked(mouse_x, mouse_y, flag)) {
					return true
				} else if (child.autoKillFocus) {
					child.killFocus()
				}
			}

			if (autoKillFocus) killFocus()
		} else if (autoKillFocus) {
			killFocus()
		}

		return false
	}

	protected open fun mouseReleasedInner(mouse_x: Double, mouse_y: Double, flag: Int): Boolean {
		return true
	}

	final override fun mouseReleased(mouse_x: Double, mouse_y: Double, flag: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (ignoreMouseEventBoundaries) return mouseReleasedInner(mouse_x, mouse_y, flag)

		for (child in children) {
			if (child.mouseReleasedChecked(mouse_x, mouse_y, flag)) {
				return true
			}
		}

		return mouseReleasedInner(mouse_x, mouse_y, flag)
	}

	fun mouseReleasedChecked(mouse_x: Double, mouse_y: Double, flag: Int): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isIgnoringMouseEventBoundaries() || withinBounds(mouse_x, mouse_y)) {
			return mouseReleased(mouse_x, mouse_y, flag)
		} else if (withinExtendedBounds(mouse_x, mouse_y)) {
			for (child in children) {
				if (child.mouseReleasedChecked(mouse_x, mouse_y, flag)) {
					return true
				}
			}
		}

		return false
	}


	protected open fun mouseDraggedInner(
		mouse_x: Double,
		mouse_y: Double,
		flag: Int,
		drag_x: Double,
		drag_y: Double
	): Boolean {
		return true
	}

	final override fun mouseDragged(mouse_x: Double, mouse_y: Double, flag: Int, drag_x: Double, drag_y: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (ignoreMouseEventBoundaries) return mouseDraggedInner(mouse_x, mouse_y, flag, drag_x, drag_y)

		for (child in children) {
			if (child.mouseDraggedChecked(mouse_x, mouse_y, flag, drag_x, drag_y)) {
				return true
			}
		}

		return mouseDraggedInner(mouse_x, mouse_y, flag, drag_x, drag_y)
	}

	fun mouseDraggedChecked(mouse_x: Double, mouse_y: Double, flag: Int, drag_x: Double, drag_y: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isIgnoringMouseEventBoundaries() || withinBounds(mouse_x, mouse_y)) {
			return mouseDragged(mouse_x, mouse_y, flag, drag_x, drag_y)
		} else if (withinExtendedBounds(mouse_x, mouse_y)) {
			for (child in children) {
				if (child.mouseDraggedChecked(mouse_x, mouse_y, flag, drag_x, drag_y)) {
					return true
				}
			}
		}

		return false
	}


	protected open fun mouseScrolledInner(mouse_x: Double, mouse_y: Double, scroll: Double): Boolean {
		return true
	}

	final override fun mouseScrolled(mouse_x: Double, mouse_y: Double, scroll: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (ignoreMouseEventBoundaries) return mouseScrolledInner(mouse_x, mouse_y, scroll)

		for (child in children) {
			if (child.mouseScrolledChecked(mouse_x, mouse_y, scroll)) {
				return true
			}
		}

		return mouseScrolledInner(mouse_x, mouse_y, scroll)
	}

	fun mouseScrolledChecked(mouse_x: Double, mouse_y: Double, scroll: Double): Boolean {
		if (!isVisible() || !acceptMouseInput) return false

		if (isIgnoringMouseEventBoundaries() || withinBounds(mouse_x, mouse_y)) {
			return mouseScrolled(mouse_x, mouse_y, scroll)
		} else if (withinExtendedBounds(mouse_x, mouse_y)) {
			for (child in children) {
				if (child.mouseScrolledChecked(mouse_x, mouse_y, scroll)) {
					return true
				}
			}
		}

		return false
	}

	protected open fun keyPressedInternal(p_94745_: Int, p_94746_: Int, p_94747_: Int): Boolean {
		return false
	}

	final override fun keyPressed(p_94745_: Int, p_94746_: Int, p_94747_: Int): Boolean {
		if (!isVisible() || !acceptKeyboardInput) return false
		if (!focusedAsParent) return false

		if (isFocused) return keyPressedInternal(p_94745_, p_94746_, p_94747_)

		for (child in children) {
			if (child.keyPressed(p_94745_, p_94746_, p_94747_)) {
				return true
			}
		}

		return false
	}

	protected open fun keyReleasedInternal(p_94750_: Int, p_94751_: Int, p_94752_: Int): Boolean {
		return false
	}

	final  override fun keyReleased(p_94750_: Int, p_94751_: Int, p_94752_: Int): Boolean {
		if (!isVisible() || !acceptKeyboardInput) return false
		if (!focusedAsParent) return false

		if (isFocused) return keyReleasedInternal(p_94750_, p_94751_, p_94752_)

		for (child in children) {
			if (child.keyReleased(p_94750_, p_94751_, p_94752_)) {
				return true
			}
		}

		return false
	}

	protected open fun charTypedInternal(p_94732_: Char, p_94733_: Int): Boolean {
		return false
	}

	final override fun charTyped(p_94732_: Char, p_94733_: Int): Boolean {
		if (!isVisible() || !acceptKeyboardInput) return false
		if (!focusedAsParent) return false

		if (isFocused) return charTypedInternal(p_94732_, p_94733_)

		for (child in children) {
			if (child.charTyped(p_94732_, p_94733_)) {
				return true
			}
		}

		return true
	}

	override fun changeFocus(p_94756_: Boolean): Boolean {
		return if (!isVisible()) false else super.changeFocus(p_94756_)
	}

	final override fun isMouseOver(
		mouse_x: Double,
		mouse_y: Double
	): Boolean { // called to check whenever we are hovering at this
		if (!isVisible() || !acceptMouseInput) return false
		if (isIgnoringMouseEventBoundaries()) return true

		val pos = localToScreen()
		return mouse_x >= pos.x && mouse_x <= pos.x + width && mouse_y >= pos.y && mouse_y + height <= pos.y
	}

	protected var tick = 0

	open fun tick() {
		tick++

		for (child in children) {
			child.tick()
		}
	}


	var isRemoved = false
		private set

	protected open fun onRemoved() {}

	fun remove() {
		if (isRemoved) {
			return
		}

		killFocus()
		screen.removePanel(this)

		for (child in children) {
			child.remove()
		}

		onRemoved()
		isRemoved = true
	}

	fun asGrid(): FlexGridPanel {
		val grid = FlexGridPanel(screen, parent, x, y, width, height)
		parent = grid
		grid.dock = dock
		dock = Dock.NONE
		grid.dockPadding = dockPadding
		grid.dockMargin = dockMargin
		setDockPadding(0f, 0f, 0f, 0f)
		setDockMargin(0f, 0f, 0f, 0f)
		return grid
	}
}
