package ru.dbotthepony.mc.otm.core

import com.mojang.math.Quaternion
import com.mojang.math.Vector3f
import net.minecraft.core.Direction
import net.minecraft.world.phys.Vec3
import kotlin.math.*

typealias Vector = Vec3

@JvmField val VECTOR_UP = Vector(0.0, 1.0, 0.0)
@JvmField val VECTOR_FORWARD = Vector(1.0, 0.0, 0.0)
@JvmField val VECTOR_BACKWARD = Vector(-1.0, 0.0, 0.0)
@JvmField val VECTOR_DOWN = Vector(0.0, -1.0, 0.0)
@JvmField val VECTOR_RIGHT = Vector(0.0, 0.0, 1.0)
@JvmField val VECTOR_LEFT = Vector(0.0, 0.0, -1.0)

fun Vector.asAngle(): Angle {
	val norm = normalize()
	return Angle(asin(norm.y), atan2(norm.x, norm.z))
}

fun Vector.asMutableAngle(): MutableAngle {
	val norm = normalize()
	return MutableAngle(asin(norm.y), atan2(norm.x, norm.z))
}

fun Vector3f.asAngle(): Angle {
	val len = length()
	val norm = Vector(x() / len, y() / len, z() / len)
	return Angle(asin(norm.y), atan2(norm.x, norm.z))
}

fun Vector3f.asMutableAngle(): MutableAngle {
	val len = length()
	val norm = Vector(x() / len, y() / len, z() / len)
	return MutableAngle(asin(norm.y), atan2(norm.x, norm.z))
}

operator fun Vector3f.unaryMinus(): Vector3f {
	setX(-x())
	setY(-y())
	setZ(-z())
	return this
}

fun Vector3f.length() = sqrt(x().toDouble() * x() + y().toDouble() * y() + z().toDouble() * z())

operator fun Vector3f.times(v: Float): Vector3f {
	setX(x() * v)
	setY(y() * v)
	setZ(z() * v)
	return this
}

operator fun Vector3f.div(v: Float): Vector3f {
	setX(x() / v)
	setY(y() / v)
	setZ(z() / v)
	return this
}

operator fun Vector3f.component1() = x()
operator fun Vector3f.component2() = y()
operator fun Vector3f.component3() = z()

operator fun Vector.plus(direction: Vector): Vector = this.add(direction)
operator fun Vector.minus(direction: Vector): Vector = this.subtract(direction)
operator fun Vector.unaryMinus(): Vector = Vector(-x, -y, -z)
operator fun Vector.component1() = x
operator fun Vector.component2() = y
operator fun Vector.component3() = z
operator fun Vector.times(v: Double): Vector = this.multiply(v, v, v)
operator fun Vector.div(v: Double): Vector {
	val inv = 1.0 / v
	return this.multiply(inv, inv, inv)
}

fun Vector.left() = asAngle().left() * this.length()
fun Vector.right() = asAngle().right() * this.length()
fun Vector.up() = asAngle().up() * this.length()
fun Vector.down() = asAngle().down() * this.length()

fun Vector.rotateAroundAxis(axis: Vector, rotation: Double): Vector {
	val cross = axis.cross(this)
	val dot = axis.dot(this)
	val cos = cos(rotation)
	val sin = sin(rotation)

	return Vector(
		this.x * cos + cross.x * sin + axis.x * dot * (1.0 - cos),
		this.y * cos + cross.y * sin + axis.y * dot * (1.0 - cos),
		this.z * cos + cross.z * sin + axis.z * dot * (1.0 - cos),
	)
}

fun Vector.rotate(angle: IAngle): Vector {
	val asMatrix = asMatrix()
	angle.rotationXYZ().multiply(asMatrix, inverse = true)
	return Vector(asMatrix[0, 0], asMatrix[1, 0], asMatrix[2, 0])
}

fun Vector.asVector3f(): Vector3f {
	return Vector3f(x.toFloat(), y.toFloat(), z.toFloat())
}

fun Quaternion(vec: Vector, rotation: Float, isDegrees: Boolean = false): Quaternion {
	return Quaternion(vec.asVector3f(), rotation, isDegrees)
}

fun Vector(x: Double): Vector {
	return Vector(x, 0.0, 0.0)
}

fun Vector(x: Double, y: Double): Vector {
	return Vector(x, y, 0.0)
}

fun Vector.rotate(q: Quaternion): Vector {
	val quaternion = Quaternion(q)
	quaternion.mul(Quaternion(x.toFloat(), y.toFloat(), z.toFloat(), 0.0f))
	val quaternion1 = Quaternion(q)
	quaternion1.conj()
	quaternion.mul(quaternion1)
	return Vector(quaternion.i().toDouble(), quaternion.j().toDouble(), quaternion.k().toDouble())
}

fun Vector.rotateAroundThis(rotation: Float, isDegrees: Boolean = false): Quaternion {
	return Quaternion(this, rotation, isDegrees)
}

fun Vector.asMatrix(): MutableMatrix {
	return MutableMatrix(3, 1).also {
		it[0, 0] = x
		it[1, 0] = y
		it[2, 0] = z
	}
}

fun Vector.asMatrix4(): MutableMatrix {
	return MutableMatrix(4, 1).also {
		it[0, 0] = x
		it[1, 0] = y
		it[2, 0] = z
	}
}

fun Direction.asAngle(): Angle {
	return when (this) {
		Direction.DOWN -> Angle.DOWN
		Direction.UP -> Angle.UP
		Direction.NORTH -> Angle.NORTH
		Direction.SOUTH -> Angle.SOUTH
		Direction.WEST -> Angle.WEST
		Direction.EAST -> Angle.EAST
	}
}

interface IAngle {
	val pitch: Double
	val yaw: Double
	val roll: Double

	fun forward() = VECTOR_FORWARD.rotate(this)
	fun left() = VECTOR_LEFT.rotate(this)
	fun right() = VECTOR_RIGHT.rotate(this)
	fun up() = VECTOR_UP.rotate(this)
	fun down() = VECTOR_DOWN.rotate(this)

	fun rotationX(): MutableMatrix {
		if (roll == 0.0) {
			return MutableMatrix(3, 3).also { it.identityFast() }
		}

		return MutableMatrix(3, 3).also {
			val s = sin(roll)
			val c = cos(roll)

			it[0, 0] = 1.0
			it[1, 1] = c
			it[1, 2] = -s
			it[2, 1] = s
			it[2, 2] = c
		}
	}

	fun rotationZ(): MutableMatrix {
		if (pitch == 0.0) {
			return MutableMatrix(3, 3).also { it.identityFast() }
		}

		return MutableMatrix(3, 3).also {
			val s = sin(pitch)
			val c = cos(pitch)

			it[0, 0] = c
			it[0, 1] = -s
			it[1, 0] = s
			it[1, 1] = c
			it[2, 2] = 1.0
		}
	}

	fun rotationY(): MutableMatrix {
		if (yaw == 0.0) {
			return MutableMatrix(3, 3).also { it.identityFast() }
		}

		return MutableMatrix(3, 3).also {
			val s = sin(yaw)
			val c = cos(yaw)

			it[0, 0] = c
			it[0, 2] = s

			it[2, 0] = -s
			it[2, 2] = c

			it[1, 1] = 1.0
		}
	}

	fun rotationXYZ(): MutableMatrix {
		if (pitch == 0.0 && yaw == 0.0 && roll == 0.0)
			return MutableMatrix(3, 3).also { it.identityFast() }

		return rotationY().multiply(rotationZ()).multiply(rotationX())
	}

	fun rotationXYZW(): MutableMatrix {
		if (pitch == 0.0 && yaw == 0.0 && roll == 0.0)
			return MutableMatrix(4, 4).also { it.identityFast() }

		return rotationXYZ().resize(4, 4).also { it[3, 3] = 1.0 }
	}
}

data class Angle(override val pitch: Double = 0.0, override val yaw: Double = 0.0, override val roll: Double = 0.0) : IAngle {
	companion object {
		const val PI_HALF = Math.PI / 2.0

		fun deg(pitch: Double, yaw: Double, roll: Double): Angle {
			return Angle(Math.toDegrees(pitch), Math.toDegrees(yaw), Math.toDegrees(roll))
		}

		@JvmField val ZERO = Angle()
		@JvmField val UP = Angle(pitch = PI / 2)
		@JvmField val DOWN = Angle(pitch = -PI / 2)
		@JvmField val NORTH = Angle()
		@JvmField val SOUTH = Angle(yaw = PI)
		@JvmField val WEST = Angle(yaw = PI / 2)
		@JvmField val EAST = Angle(yaw = -PI / 2)
	}
}

data class MutableAngle(override var pitch: Double = 0.0, override var yaw: Double = 0.0, override var roll: Double = 0.0) : IAngle {
	companion object {
		fun deg(pitch: Double, yaw: Double, roll: Double): MutableAngle {
			return MutableAngle(Math.toDegrees(pitch), Math.toDegrees(yaw), Math.toDegrees(roll))
		}
	}
}
